import multiprocessing as mp
import queue
import time
from typing import Callable, Iterable, Iterator, List, Optional

from .handler import BaseTaskHandler
from .logger import log
from .metrics.timer import timeit
from .task import BaseTask, StopTask


def batches(
        elements: Iterable,
        batch_size: int,
        timeout: float,
) -> Iterator[List]:
    batch = []
    timeout_end = time.monotonic() + timeout

    for elem in elements:
        if elem:
            batch.append(elem)
        if time.monotonic() >= timeout_end or len(batch) == batch_size:
            if batch:
                yield batch
                batch = []
            timeout_end = time.monotonic() + timeout

    if batch:
        yield batch


def batches_with_lock(batches_gen: Iterator[List], lock: mp.Lock) -> Iterator[List]:
    while True:
        lock.acquire()
        try:
            batch = next(batches_gen)
        except StopIteration:
            return
        finally:
            lock.release()
        yield batch


class Worker:
    """Обертка над классом BaseTaskHandler.

    Достает из входной очереди задачу, пропускает ее через обработчика task_handler и кладет ее
    в выходную очередь.
    """

    def __init__(
            self,
            queue_in: mp.Queue,
            queue_out: mp.Queue,
            task_handler: BaseTaskHandler,
            handle_condition: Callable[[BaseTask], bool],
            batch_size: int,
            batch_timeout: float,
            batch_lock: Optional[mp.RLock],
            step_number: int,
    ):
        self.queue_in = queue_in
        self.queue_out = queue_out
        self.task_handler = task_handler
        self.handle_condition = handle_condition
        self.name = task_handler.__class__.__name__
        self.step_name = self.task_handler.get_step_name(step_number)
        self._batch_size = batch_size
        self._batch_timeout = batch_timeout
        self._batch_lock = batch_lock
        self._stop_task: BaseTask = None  # noqa

    def _start(self):
        """Runs something huge (e.g. model) in child process."""
        self.task_handler.on_start()

    def _tasks(self) -> Iterator[Optional[BaseTask]]:
        """Provides suitable for processing tasks."""
        while True:
            try:
                task: BaseTask = self.queue_in.get(block=False)
            except queue.Empty:
                # returns control
                yield
                time.sleep(0.001)
                continue
            if isinstance(task, StopTask):
                self._stop_task = task
                break
            log.debug(f'[{self.name}] Have message')
            task.metrics.stop_transfer_timer(self.step_name)
            # dont't pass an expired task to the next steps
            if task.is_expired():
                log.debug(f'[{self.name}] Task expired. Skip: {task}')
                continue
            # don't process unsuitable tasks
            if not self.handle_condition(task):
                self._post_handle(task)
                continue
            yield task

    def _tasks_batches(self) -> Iterator[Optional[List[BaseTask]]]:
        if self._batch_size == 1:
            # pseudo batching
            for task in self._tasks():
                if task:
                    yield [task]
        else:
            tasks_batches: Iterator[List[BaseTask]] = batches(
                self._tasks(),
                batch_size=self._batch_size,
                timeout=self._batch_timeout,
            )
            if self._batch_lock:
                # to take a queue_in lock for the duration of batch filling time
                tasks_batches = batches_with_lock(tasks_batches, self._batch_lock)

            while True:
                try:
                    with timeit() as timer:
                        tasks_batch = next(tasks_batches)
                    tasks_batch[0].metrics.batch_times.add(self.step_name, timer.seconds)
                    tasks_batch[0].metrics.batch_sizes.add(self.step_name, len(tasks_batch))
                except StopIteration:
                    return
                yield tasks_batch

    def _post_handle(self, task: BaseTask):
        task.metrics.start_transfer_timer(self.step_name)
        self.queue_out.put(task)

    def loop(self, pid: int, start_barrier: mp.Barrier):
        """Main worker loop.

        The code below is executed in a new process.
        """
        log.info(f'[Worker] initialising handler {self.name}')
        self._start()
        log.info(f'[Worker] handler {self.name} ok, waiting for others to start')
        start_barrier.wait()

        log.info(f'[Worker] handler {self.name} ok, starting loop')

        for tasks_batch in self._tasks_batches():
            with timeit() as timer:
                self.task_handler.handle(*tasks_batch)

            for task in tasks_batch:
                task.metrics.handle_times.add(self.step_name, timer.seconds)
                self._post_handle(task)

        if self._stop_task:
            self.queue_out.put(self._stop_task)



import time

import pytest

from aqueduct.worker import batches


def long_elements_producer():
    """Long non-blocking producer of elements."""
    tic_elements = {1: 1, 4: 2, 6: 3, 9: 4}

    for tic in range(10):
        if tic in tic_elements:
            yield tic_elements[tic]
        else:
            yield
        time.sleep(0.01)


@pytest.mark.parametrize('elements, batch_size, result', [
    ([1, 2, 3], 1, [[1], [2], [3]]),
    ([1, 2, 3], 2, [[1, 2], [3]]),
    ([1, 2, 3], 4, [[1, 2, 3]]),
])
def test_batches_different_batch_sizes(elements, batch_size, result):
    assert list(batches(elements, batch_size, 1.)) == result


@pytest.mark.parametrize('timeout, result', [
    (0.01, [[1], [2], [3], [4]]),
    (0.03, [[1], [2, 3], [4]]),
])
def test_batches_different_timeouts(timeout, result):
    assert list(
        batches(long_elements_producer(), 4, timeout)
    ) == result



from .task import BaseTask


class BaseTaskHandler:
    """Базовый класс для ваших хендлеров.
       Используется для логики которой не требуется модель и специальная инициализация.
       Если же нужна модель, которая долго стартует и занимает память, то используйте ModelTaskHandler()."""

    @classmethod
    def get_step_name(cls, step_number: int) -> str:
        return f'step{step_number}_{cls.__name__}'

    def on_start(self):
        """Вызывается при старте в дочернем процессе. Чтобы модель занимала память только в подпроцессе,
        а не в родителе."""
        pass

    def handle(self, *tasks: BaseTask):
        """Вызывается при получении из очереди задания."""
        raise NotImplementedError



import asyncio
import sys

from aqueduct.flow import Flow
from aqueduct.logger import log

from aiohttp import web


AQUEDUCT_FLOW_NAMES = 'aqueduct_flow_names'
FLOW_NAME = 'flow'
FLOWS_OBSERVER = 'aqueduct_flows_observer'


async def observe_flows(app: web.Application, check_interval: float = 1.):
    flows = {name: app[name] for name in app[AQUEDUCT_FLOW_NAMES]}
    while True:
        for flow_name, flow in flows.items():
            if not flow.is_running:
                log.info(f'Flow {flow_name} is not running, application will be stopped')
                sys.exit(1)
        await asyncio.sleep(check_interval)


async def run_flows_observer(app):
    app[FLOWS_OBSERVER] = asyncio.create_task(observe_flows(app))


async def stop_flows_observer(app):
    app[FLOWS_OBSERVER].cancel()


async def stop_flows(app):
    flows = [app[name] for name in app[AQUEDUCT_FLOW_NAMES]]
    for flow in flows:  # type: Flow
        await flow.stop()


class AppIntegrator:
    """Adds to app flows and actions to manage flows and app itself."""
    def __init__(self, app: web.Application, exit_on_fail: bool = True):
        if AQUEDUCT_FLOW_NAMES in app:
            raise RuntimeError('AppIntegrator can be created only once. Reuse existing AppIntegrator.')
        self._app = app
        self._app[AQUEDUCT_FLOW_NAMES] = []
        if exit_on_fail:
            self._app.on_startup.append(run_flows_observer)
            self._app.on_shutdown.append(stop_flows_observer)
        self._app.on_shutdown.append(stop_flows)

    def add_flow(self, flow: Flow, flow_name: str = FLOW_NAME, with_start: bool = True):
        if flow_name in self._app:
            raise RuntimeError(f'Flow with name "{flow_name}" already exists in app')
        if with_start:
            flow.start()
        self._app[AQUEDUCT_FLOW_NAMES].append(flow_name)
        self._app[flow_name] = flow



from multiprocessing import Process, Queue
from unittest.mock import patch

import numpy as np
import pytest

from aqueduct.shm import (
    NPArraySharedData,
    SharedMemoryWrapper,
)
from tests.unit.conftest import Task


class ArrayFieldTask(Task):
    def __init__(self, array: np.array, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.array = array


def update_sh_array(q_in: Queue, q_out: Queue):
    array_sh_data, value = q_in.get()
    sh_array = array_sh_data.get_data()
    sh_array[0] = value
    q_out.put((None, None))
    # дожидаемся изменения массива в главном процессе
    _, new_value = q_in.get()
    # теперь первый элемент массива равен new_value
    # сама проверка делается в родительском процессе, потому что исключение из дочернего не прокидывается
    # в родительский процесс автоматически
    q_out.put((sh_array[0] == new_value, f'array elem {sh_array[0]} is not equal to value {new_value}'))


def update_ref_counter(q_in: Queue, q_out: Queue):
    array_sh_data, value = q_in.get()
    shm_wrapper: SharedMemoryWrapper = array_sh_data.shm_wrapper
    ref_counter = shm_wrapper._ref_counter  # noqa
    ref_counter[0] = value
    q_out.put((None, None))
    # дожидаемся изменения ref_counter в главном процессе
    _, new_value = q_in.get()
    # теперь ref_counter равен new_value
    q_out.put((ref_counter[0] == new_value, f'ref_count {ref_counter[0]} is not equal to value {new_value}'))


def update_task_field(q_in: Queue, q_out: Queue):
    task, field_name, value = q_in.get()
    array = getattr(task, field_name)
    array[0] = value
    q_out.put(None)


def close_access_to_shm(q_in: Queue, q_out: Queue):
    array_sh_data: NPArraySharedData = q_in.get()
    # closes access to shm
    array_sh_data.shm_wrapper.shm.close()
    q_out.put(None)
    q_in.get()
    # error when trying to use shm
    try:
        array_sh_data.get_data()
    except ValueError as e:
        q_out.put(e)


@pytest.fixture
def task(array: np.ndarray):
    return ArrayFieldTask(array)


def handle_task(q_in: Queue, q_out: Queue):
    # была присоединена разделяемая память
    task = q_in.get()
    q_out.put(None)

    # объект, который ссылался на резделяемую память, должен быть удален gc
    task = q_in.get()
    q_out.put(None)

    q_in.get()


class TestNPArraySharedData:
    def test_access_to_data(self, array_sh_data: NPArraySharedData):
        sh_array = array_sh_data.get_data()

        q_in, q_out = Queue(), Queue()
        p = Process(target=update_sh_array, args=(q_out, q_in))
        p.start()

        value = sh_array[0] + 1
        assert sh_array[0] != value
        q_out.put((array_sh_data, value))
        # дожидаемся изменения массива в дочернем процессе
        _, _ = q_in.get()
        # теперь первый элемент массива равен value
        assert sh_array[0] == value
        value += 1
        sh_array[0] = value
        q_out.put((None, value))
        assert_expr, msg = q_in.get()
        assert assert_expr, msg

        p.join()

    def test_access_to_ref_counter(self, array_sh_data: NPArraySharedData):
        ref_counter = array_sh_data.shm_wrapper._ref_counter

        q_in, q_out = Queue(), Queue()
        p = Process(target=update_ref_counter, args=(q_out, q_in))
        p.start()

        value = 4
        q_out.put((array_sh_data, value))
        # дожидаемся изменения массива в дочернем процессе
        _, _ = q_in.get()
        # теперь ref_counter равен value
        assert ref_counter[0] == value
        value = 2
        ref_counter[0] = value
        q_out.put((None, value))
        assert_expr, msg = q_in.get()
        assert assert_expr, msg

        p.join()

    # todo добавить тесты на подсчет ссылок и удаление объектов, unlink памяти, переприсваивание массива, SharedData
    def test_access_to_shm_after_close(self, array_sh_data: NPArraySharedData):
        q_in, q_out = Queue(), Queue()
        p = Process(target=close_access_to_shm, args=(q_out, q_in))
        p.start()

        # creates array successfully in parent process
        sh_array = array_sh_data.get_data()
        sh_array[0] = 42
        # send shm data to child process
        q_out.put(array_sh_data)
        # access to shm were closed in child process
        q_in.get()
        # creates array successfully in parent process
        sh_array = array_sh_data.get_data()
        assert sh_array[0] == 42
        q_out.put(None)
        # error in child process when trying to use shm
        exc = q_in.get()
        assert isinstance(exc, ValueError)

        p.join()

        # creates array successfully after child process is finished
        sh_array = array_sh_data.get_data()
        assert sh_array[0] == 42


class TestSharedFieldsMixin:
    def test_pickle_task(self, task: ArrayFieldTask):
        field_name = 'array'

        state = task.__getstate__()
        assert field_name in state
        assert state['_shared_fields'] == {}

        # если поле расшарено, то оно не пиклится, вместо него пиклится мета информация
        task.share_value(field_name)
        state = task.__getstate__()
        assert field_name not in state
        assert field_name in state['_shared_fields']

    def test_update_task_field(self, task: ArrayFieldTask):
        task.share_value('array')

        q_in, q_out = Queue(), Queue()
        p = Process(target=update_task_field, args=(q_out, q_in))
        p.start()

        value = task.array[0] + 1
        q_out.put((task, 'array', value))
        # дожидаемся изменения массива в дочернем процессе
        _ = q_in.get()
        # теперь первый элемент массива равен value
        assert task.array[0] == value

        p.join()

    def test_auto_change_ref_counter(self, task: ArrayFieldTask):
        task.share_value('array')
        shf_info = task._shared_fields['array']
        ref_counter = shf_info.shm_wrapper._ref_counter
        # при присоединении памяти был увеличен счетчик ссылок
        assert ref_counter[0] == 1

        q_in, q_out = Queue(), Queue()
        p = Process(target=handle_task, args=(q_out, q_in))
        p.start()

        q_out.put(task)
        q_in.get()
        # после присоединения области памяти дочерним процессом был увеличен счетчик ссылок
        assert ref_counter[0] == 2

        q_out.put(None)
        q_in.get()
        # после работы gc в дочернем процессе счетчик ссылок был уменьшен
        assert ref_counter[0] == 1

        q_out.put(None)
        p.join()

        task.array = None
        with patch.object(shf_info.shm_wrapper.shm, 'unlink') as mocked:
            shf_info, ref_counter = None, None
            # после того, как не остается ссылок на разделяемую память, она освобождается
            assert mocked.called is True

    def test_save_shared_value_directly(self, task: ArrayFieldTask):
        # НЕ рекомендуемый способ расшарить значение поля, но так тоже можно
        field_name = 'array'

        shared_array = NPArraySharedData.create_from_data(task.array)
        task.array = shared_array
        assert field_name in task._shared_fields
        assert task._shared_fields[field_name].shm_wrapper._ref_counter[0] == 1



import asyncio
import multiprocessing as mp
import os
import queue
import signal
import sys
from enum import Enum
from functools import cached_property
from functools import reduce
from multiprocessing import Barrier
from threading import BrokenBarrierError
from typing import Callable, Dict, List, Optional, Union

from multiprocessing.resource_tracker import _resource_tracker  # noqa

from .exceptions import FlowError, NotRunningError
from .handler import BaseTaskHandler
from .logger import log
from .metrics import MAIN_PROCESS, MetricsTypes
from .metrics.collect import Collector, TasksStats
from .metrics.export import Exporter
from .metrics.manager import get_metrics_manager
from .metrics.timer import timeit
from .multiprocessing import (
    ProcessContext,
    ProcessExitedException,
    ProcessRaisedException,
    start_processes,
)
from .task import BaseTask, StopTask
from .worker import Worker

# just for using common ResourceTracker in main and child processes and avoiding
# unnecessary shared memory resource_tracker "No such file or directory" warnings
_resource_tracker.ensure_running()


def _check_env():
    if not sys.version_info >= (3, 8):
        raise RuntimeError('Requires python 3.8 or higher to use multiprocessing.shared_memory')


class FlowStep:

    def __init__(
            self,
            handler: BaseTaskHandler,
            handle_condition: Callable[[BaseTask], bool] = lambda x: True,
            nprocs: int = 1,
            batch_size: int = 1,
            batch_timeout: float = 0.5,
    ):
        _check_env()
        self.handler = handler
        self.handle_condition = handle_condition
        self.nprocs = nprocs
        self.batch_size = batch_size
        self.batch_timeout = batch_timeout


class FlowState(Enum):
    RUNNING = 'running'
    STOPPING = 'stopping'
    STOPPED = 'stopped'


class Flow:

    def __init__(
            self,
            *steps: Union[FlowStep, BaseTaskHandler],
            metrics_enabled: bool = True,
            metrics_collector: Collector = None,
            metrics_exporter: Exporter = None,
            queue_size: int = 20,
    ):
        _check_env()

        self._steps: List[FlowStep] = [step if isinstance(step, FlowStep)
                                       else FlowStep(step) for step in steps]
        self._contexts: Dict[BaseTaskHandler, ProcessContext] = {}
        self._queues: List[mp.Queue] = []
        self._task_futures: Dict[str, asyncio.Future] = {}
        self._queue_size: int = queue_size

        if not metrics_enabled:
            log.warn('Metrics collecting is disabled')
            metrics_collector = Collector(collectible_metrics=[])
        self._metrics_manager = get_metrics_manager(metrics_collector, metrics_exporter)

        self._state: FlowState = None  # noqa
        self._tasks: List[asyncio.Future] = []

    @property
    def state(self):
        return self._state

    @property
    def is_running(self) -> bool:
        return self._state == FlowState.RUNNING

    def start(self, timeout: Optional[int] = None):
        """
        Starts Flow and waits for all subprocesses to initialize.

        raising FlowError if 'timeout' is set and some handler was not able to initialize in time.
        """
        log.info('Flow is starting')
        self._run_steps(timeout)
        self._run_tasks()
        self._state = FlowState.RUNNING
        log.info('Flow was started')

    async def process(self, task: BaseTask, timeout_sec: float = 5.) -> bool:
        """
        Starts processing task through pipeline and returns True if all was ok.
        The result is saved in the source task
        """

        if not self.is_running:
            raise NotRunningError

        with timeit() as timer:

            future = asyncio.Future()
            self._task_futures[task.task_id] = future

            task.set_timeout(timeout_sec)
            task.metrics.start_transfer_timer(MAIN_PROCESS)
            while True:
                try:
                    self._queues[0].put(task, block=False)
                except queue.Full:
                    await asyncio.sleep(0.001)
                else:
                    break

            tasks_stats = TasksStats()
            try:
                finished_task: BaseTask = await asyncio.wait_for(future, timeout=timeout_sec)
            # todo is it correct to hide a specific error behind a general FlowError?
            except asyncio.TimeoutError:
                tasks_stats.timeout += 1
                raise FlowError('Task timeout error')
            except asyncio.CancelledError:
                tasks_stats.cancel += 1
                raise FlowError('Task was cancelled')
            else:
                tasks_stats.complete += 1
            finally:
                self._metrics_manager.collector.add_tasks_stats(tasks_stats)
                del self._task_futures[task.task_id]

        finished_task.metrics.handle_times.add('total', timer.seconds)

        if self.need_collect_task_timers:
            self._metrics_manager.collector.add_task_metrics(finished_task.metrics)

        task.update(finished_task)

        return True

    async def stop(self, graceful: bool = True):
        if not self.is_running:
            log.info('Flow is not running')
            return

        self._state = FlowState.STOPPING
        log.info('Flow is stopping')

        if graceful:
            self._queues[0].put(StopTask())
            await asyncio.sleep(3)

        self._metrics_manager.stop()
        for task in self._tasks:
            task.cancel()
        for task in self._task_futures.values():
            task.cancel()

        self._join_context(self._processes_context)

        self._state = FlowState.STOPPED
        log.info('Flow was stopped')

    @cached_property
    def need_collect_task_timers(self) -> bool:
        return self._metrics_manager.collector.is_collectible(MetricsTypes.TASK_TIMERS)

    def _run_steps(self, timeout: Optional[int]):
        self._queues.append(mp.Queue(self._queue_size))

        total_procs = reduce(lambda a, b: a + b.nprocs, self._steps, 0)
        # also count main process
        total_procs += 1
        start_barrier = Barrier(total_procs)

        for step_number, step in enumerate(self._steps, 1):
            self._queues.append(mp.Queue(self._queue_size))
            worker_curr = Worker(
                self._queues[-2],
                self._queues[-1],
                step.handler,
                step.handle_condition,
                step.batch_size,
                step.batch_timeout,
                mp.RLock() if step.nprocs > 1 and step.batch_size > 1 else None,
                step_number,
            )
            self._contexts[step.handler] = start_processes(
                worker_curr.loop,
                nprocs=step.nprocs, join=False, daemon=True, start_method='fork',
                args=(start_barrier,),
            )
            log.info(f'Created step {step.handler}, '
                     f'queue_in: {self._queues[-2]}, queue_out:{self._queues[-1]}')

        try:
            log.info('Waiting for all workers to startup...')
            start_barrier.wait(timeout)
        except BrokenBarrierError:
            raise TimeoutError('Starting timeout expired')

    def _run_tasks(self):
        self._tasks.append(asyncio.ensure_future(self._fetch_processed()))
        self._tasks.append(asyncio.ensure_future(self._check_is_alive()))

        self._metrics_manager.start(queues_info=self._get_queues_info())

    def _get_queues_info(self) -> Dict[mp.Queue, str]:
        """Returns queues between Step handlers and its names.

        The queue name consists of two handler names that are connected by this queue.
        """
        def step_names(handlers):
            from_step = MAIN_PROCESS
            for step_number, handler in enumerate(handlers, 1):
                to_step = handler.get_step_name(step_number)
                yield from_step, to_step
                from_step = to_step
            yield from_step, MAIN_PROCESS

        return {queue_: f'from_{from_}_to_{to}'
                for queue_, (from_, to) in zip(self._queues, step_names(self._contexts))}

    async def _fetch_processed(self):
        while True:
            try:
                task: BaseTask = self._queues[-1].get(block=False)
                task.metrics.stop_transfer_timer(MAIN_PROCESS)
            except queue.Empty:
                await asyncio.sleep(0.001)
                continue
            if task.task_id in self._task_futures:
                future = self._task_futures[task.task_id]
                if not future.cancelled() and not future.done():
                    future.set_result(task)

    async def _check_is_alive(self, sleep_sec: float = 1.):
        """Checks that all child processes are alive.

        If at least one process is not alive, it stops Flow.
        """
        while True:
            for handler, context in self._contexts.items():
                for proc in context.processes:
                    if not proc.is_alive():
                        if self.is_running:
                            log.error('The process %s for %s handler is dead',
                                      proc.pid, handler.__class__.__name__)
                            await self.stop(graceful=False)
            await asyncio.sleep(sleep_sec)

    @staticmethod
    def _join_context(context: ProcessContext, timeout_sec: float = 0.01):
        try:
            context.join(timeout_sec)
        except (ProcessExitedException, ProcessRaisedException):
            pass

        for p in context.processes:
            if p.is_alive():
                os.kill(p.pid, signal.SIGKILL)

    @property
    def _processes_context(self) -> ProcessContext:
        processes, error_queues = [], []
        for context in self._contexts.values():
            processes.extend(context.processes)
            error_queues.extend(context.error_queues)
        return ProcessContext(processes=processes, error_queues=error_queues)



"""Предоставляет инструменты для управления данными в разделяемой памяти.

В shared_memory есть особенность - даже после close() в другом процессе, где была прилинкована память,
при выходе из процесса вызывается unlink(). Как следствие - этот кусок памяти не может быть больше
нигде использован, т.к. возвращен ОС. Причина - в resource_tracker'е, который для каждого не дочернего
процесса - отдельный.

Обойти эту особенность можно, явно исключив из наблюдения трекера, кусок памяти:
from multiprocessing.resource_tracker import unregister
unregister(f'/{shm.name}', 'shared_memory')

Также этой проблемы нет, если процесс, использующий память, дочерний (т.к. там тот же самый трекер ресурсов)
https://bugs.python.org/issue39959
"""
import multiprocessing as mp
import warnings
from abc import ABC, abstractmethod
try:
    from multiprocessing import shared_memory
except ImportError as e:
    raise ImportError('shared_memory module is available since python3.8')
from typing import Any, Dict, Tuple

import numpy as np

from .exceptions import BadReferenceCount


class SharedMemoryWrapper:
    """Управляет разделяемой памятью.

    Предоставляет память нужного размера, следит за количеством ссылок на память, при отсутствии
    ссылок на память возвращает память ОС.
    """
    def __init__(self, size: int, shm: shared_memory.SharedMemory = None, shm_name: str = None):
        # поле _size нужно, чтобы найти в разделяемой памяти ref_counter
        self._size = size
        self._shm: shared_memory.SharedMemory = None  # noqa
        self._ref_counter: np.ndarray = None  # noqa

        if shm is None and shm_name is None:
            # 4 байта на счетчик ссылок (int32)
            shm = shared_memory.SharedMemory(create=True, size=self._size + 4)
        self._attach_shm(shm, shm_name)

        # поле _shm_name нужно для десериализации
        self._shm_name = self._shm.name

    @property
    def shm(self) -> shared_memory.SharedMemory:
        return self._shm

    @property
    def ref_count(self) -> int:
        return self._ref_counter[0]

    def _attach_shm(self, shm: shared_memory.SharedMemory = None, shm_name: str = None):
        if shm is None:
            shm = shared_memory.SharedMemory(shm_name)
        self._shm = shm
        self._ref_counter = self._get_ref_counter()
        # нужно только для процесса, выделившего разделяемую память
        if self.ref_count == 0:
            self._update_ref_counter()

    def _update_ref_counter(self, incr: bool = True):
        """Обновляет количество ссылок на используемую область разделяемой памяти."""
        if incr:
            self._ref_counter[0] += 1
        else:
            if self._ref_counter[0] == 0:
                raise BadReferenceCount('Reference count should not be less than 0')
            self._ref_counter[0] -= 1

    def _get_ref_counter(self) -> np.ndarray:
        # todo попробовать сделать ссылку на целое число, а не np.array
        return np.frombuffer(self._shm.buf, dtype=np.int32, offset=self._size, count=1)

    def _release_shm(self):
        # if shm is already closed (for example manually). Unexpected behavior
        if self._shm.buf is None:
            # todo log as warning? ref_counter would be wrong -> memory leak
            return

        self._update_ref_counter(False)
        ref_count = self._ref_counter[0]
        self._shm.close()
        if ref_count == 0:
            self._shm.unlink()

    def __getstate__(self) -> dict:
        # счетчик ссылок увеличивается здесь, т.к. если объект пересылается долго (долго находится в очереди),
        # gc удаляет объект и память возвращается ОС до того, как объект будет вычитан из очереди
        # из-за этого при вычитывании объекта и попытке присоединить разделяемую память возникает ошибка
        self._update_ref_counter()
        state = {k: v for k, v in self.__dict__.items() if k not in ('_shm', '_ref_counter')}
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        self._attach_shm(shm_name=self._shm_name)

    def __del__(self):
        self._release_shm()


class SharedData(ABC):
    """Управляет данными в разделяемой памяти."""
    def __init__(self, shm_wrapper: SharedMemoryWrapper):
        self.shm_wrapper = shm_wrapper

    @abstractmethod
    def get_data(self) -> Any:
        """Возвращает ссылку на данные, лежащие в разделяемой памяти."""

    @classmethod
    @abstractmethod
    def create_from_data(cls, data: Any) -> 'SharedData':
        """Копирует данные в разделяемую память и возвращает экземпляр класса."""


class NPArraySharedData(SharedData):
    def __init__(self, shm_wrapper: SharedMemoryWrapper, shape: Tuple[int, ...], dtype: np.dtype):
        self._shape = shape
        self._dtype = dtype
        super().__init__(shm_wrapper)

    def get_data(self) -> np.ndarray:
        if self.shm_wrapper.shm.buf is None:
            raise ValueError('No shared memory buffer')
        return np.ndarray(self._shape, dtype=self._dtype, buffer=self.shm_wrapper.shm.buf)

    @classmethod
    def create_from_data(cls, data: np.ndarray) -> 'NPArraySharedData':
        shm_wrapper = SharedMemoryWrapper(data.nbytes)
        shared_data = cls(shm_wrapper, data.shape, data.dtype)
        array = shared_data.get_data()
        array[:] = data[:]
        return shared_data


class SharedFieldsMixin:
    """Позволяет заменить значение поля экземпляра класса на его копию в разделяемой памяти."""
    def __init__(self, *args, **kwargs):
        self._shared_fields: Dict[str, SharedData] = {}
        super().__init__(*args, **kwargs)

    def share_value(self, field_name: str):
        """Заменяет значение поля на его копию в разделяемой памяти.

        Все последующие изменения значения будут производиться в разделяемой памяти, т.е. эти изменения
        будут отражаться во всех объектах, которые ссылаются на эту же область памяти.
        В случае замены значения необходимо заново вызвать данный метод, чтобы значение попало в
        разделяемую память, при этом будет выделена новая область разделяемой памяти.
        """
        if field_name not in self.__dict__:
            raise ValueError('Unknown field name')
        if field_name in self._shared_fields:
            raise ValueError(f'The {field_name} field value is already shared')

        value = getattr(self, field_name)
        shared_value = self._get_shared_value(value)
        self._set_shared_value(field_name, shared_value)

    @staticmethod
    def _get_shared_value(value: Any) -> SharedData:
        if isinstance(value, np.ndarray):
            return NPArraySharedData.create_from_data(value)
        raise ValueError(f'Type {type(value)} cannot be shared')

    def _set_shared_value(self, field_name: str, shared_value: SharedData):
        self.__dict__[field_name] = shared_value.get_data()
        self._shared_fields[field_name] = shared_value

    def __setattr__(self, key, value):
        if isinstance(value, SharedData):
            self._set_shared_value(key, value)
        else:
            super().__setattr__(key, value)
            if key in self._shared_fields:
                self._shared_fields.pop(key)
                warnings.warn(f'The {key} field value is not shared anymore')

    def __getstate__(self) -> dict:
        state = {k: v for k, v in self.__dict__.items() if k not in self._shared_fields}
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        for f_name, shared_value in self._shared_fields.items():
            self.__dict__[f_name] = shared_value.get_data()



"""Allows you to send requests with certain rps.

This tools is written by Vladislav Kassym in Avito.
"""

import asyncio
import math
from typing import (
    Awaitable,
    Callable,
    Iterable,
    Optional,
    Sized,
    Tuple,
)

try:
    from tqdm.auto import tqdm
except ImportError:
    tqdm = None


class Monitor:
    def __init__(self, rps: float, connections: int, tqdm=None):
        self._rps = rps
        self._connections = connections
        self._tqdm = tqdm
        if self._tqdm is None:
            return
        self._pbar_total = tqdm(desc='Total', total=0)
        self._pbar_rps = tqdm(desc='RPS', total=0)
        self._pbar_retry = tqdm(desc='Retry', total=0)

    def __del__(self):
        if self._tqdm is None:
            return
        for pbar in (self._pbar_total, self._pbar_rps, self._pbar_retry):
            pbar.refresh()
            pbar.close()

    def add(self, num: int):
        if self._tqdm is None:
            return
        self._pbar_total.total += num
        self._pbar_total.refresh()

    def update(self, retry: bool = False):
        if self._tqdm is None:
            return
        self._pbar_rps.update()
        if retry:
            self._pbar_retry.update()
        else:
            self._pbar_total.update()


class Pulemet:
    """Класс для контроля скорости выполнения корутин."""

    def __init__(
        self,
        rps: float = 0.1,
        max_connections: Optional[int] = None,
        name: str = '',
        pbar: Optional[bool] = False,
    ):
        """
        Инициализация.

        Args:
            rps: количество запросов в секунду
            max_connections: максимальное количество соединений, которые одновременно ждут ответа
            name: имя пулемета
            pbar: progress bar из tqdm. Пример: from tqdm.auto import tqdm; Pulemet(pbar=tqdm())
            pbar_retry: progress bar из tqdm. Пример: from tqdm.auto import tqdm; Pulemet(pbar=tqdm())
        """
        self._rps_min, self._rps_max = 5, 10
        self._rps, self._burst = self._get_rps_and_burst(rps)

        self._max_connections = max_connections
        if max_connections is None:
            self._max_connections = int(self._rps * 5) + 10  # 5 секунд копим, 10 прост
        self._name = name

        self._semaphore_time = asyncio.BoundedSemaphore(value=math.ceil(self._burst))
        self._semaphore_connections = asyncio.Semaphore(value=self._max_connections)

        self._timer_task = asyncio.ensure_future(self._timer())

        if pbar is None:
            pbar = False
        elif not isinstance(pbar, bool):
            pbar = True
        self._pbar = Monitor(
            rps=rps, connections=self._max_connections, tqdm=tqdm if pbar else None
        )

    def __del__(self):
        self._timer_task.cancel()

    def process(self, coros: [Iterable[Awaitable], Sized]) -> [Iterable[Awaitable], Sized]:
        """
        Запускает _wrap_coro для всех корутин в списке.

        Args:
            coros: Список корутин, которые необходимо ограничить по скорости исполнения

        Returns:
            Новый список корутин

        """
        self._pbar.add(num=len(coros))
        res = [self._wrap_coro(coro) for coro in coros]

        return res

    def process_funcs(
        self,
        coro_func: Callable[..., Awaitable],
        coros_kwargs: [Iterable[dict], Sized],
        exceptions: Tuple[BaseException, ...],
        exceptions_max: Optional[int] = None,
    ) -> [Iterable[Awaitable], Sized]:
        """
        Запускает _wrap_coro для всех корутин в списке.

        Args:
            coros: Список корутин, которые необходимо ограничить по скорости исполнения

        Returns:
            Новый список корутин

        """
        if hasattr(coros_kwargs, '__len__'):
            self._pbar.add(num=len(coros_kwargs))
        res = [
            self._warp_coro_func(
                coro_func=coro_func,
                exceptions=exceptions,
                exceptions_max=exceptions_max,
                **coro_kwargs,
            )
            for coro_kwargs in coros_kwargs
        ]

        return res

    def _get_rps_and_burst(self, rps: float):
        if rps <= self._rps_max:
            rps_target, burst = rps, 1
        else:
            burst_max = int(rps / self._rps_min)
            burst_min = math.ceil(rps / self._rps_max)
            res = []
            for burst in range(burst_min, burst_max + 1):
                rps_target = round(rps / burst, 2)
                prec = abs(rps - rps_target * burst) / rps * 100
                res.append((burst, rps_target, prec))
                if prec < 0.01:
                    break
            burst, rps_target, prec = sorted(res, key=lambda x: x[2])[0]

        return rps_target, burst

    async def _timer(self):
        """Освобождает семафор на исполнение корутины `burst` раз в `1 / rps` секунд."""
        while True:
            await asyncio.sleep(1 / self._rps)
            for ind in range(self._burst):
                try:
                    self._semaphore_time.release()
                except ValueError:
                    continue

    async def _wrap_coro(self, coro: Awaitable, update: bool = True) -> Awaitable:
        """
        Ждем освобождения семафоров на rps и количество соединений и запускаем корутину.

        Args:
            coro: корутина, запуск которой необходимо ограничить по rps

        Returns:

        """
        async with self._semaphore_connections:
            await self._semaphore_time.acquire()
            if update:
                self._pbar.update(retry=False)
            result = await coro

        return result

    async def _warp_coro_func(
        self,
        coro_func: Callable[..., Awaitable],
        exceptions: Tuple[BaseException, ...],
        exceptions_max: Optional[int] = None,
        **coro_kwargs,
    ) -> Awaitable:
        cnt = 0
        while True:
            coro = coro_func(**coro_kwargs)
            try:
                coro = await self._wrap_coro(coro, update=False)
                self._pbar.update(retry=False)
                return coro
            except exceptions as exc:
                cnt += 1
                self._pbar.update(retry=True)
                if exceptions_max is not None and cnt == exceptions_max:
                    raise exc
                continue
